/*
    $(CC) -I/usr/include/libxml2 -I. -L ../../lib -l xml2 tv_scan.c dvb.c parse_dvb.c xml.c -o tv_scan


*/
//#define DEBUG 1

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include <errno.h>
#include <signal.h>

#include "dvb.h"
#include "parse_dvb.h"
#ifdef DEBUG
#include "dump_dvb.h"
#endif
#include "xml.h"

//----------------------------------------------------------------------------
int debug = 0;
int timeout_flag = 0;

int num_transponders = 0;
transponder_type transponder[10];
int fake_lcn = 900;

//----------------------------------------------------------------------------
static char FRONTEND_DEV[80];
static char DEMUX_DEV[80];
static char DVR_DEV[80];

//----------------------------------------------------------------------------

static void do_abort(int x)
{
    (void)x;
    printf("\ndo_abort()\n");
    timeout_flag = 1;
}

static char *usage_str =
    "\nusage: tv_scan [options]\n"
      "         -a <adapter>      :            [0] select adapter number\n"
      "         -f <frontend>     :            [0] select frontend number\n"
      "         -d <demux>        :            [0] select demux number\n"
      "         -c <channel_list> : [channels.xml] channel list\n"
      "         -s                :            [0] build channel list for all muxes\n"
      "         -e <frequency>    :    [505833333] initial frequency for scan\n"
      "         -t                :            [0] tune to indicated channel\n"
      "         -p                :            [0] enable promiscuous mode\n"
      "         -l <lcn>          :            [1] select channel by lcn\n"
      "         -i <service_id>   :       [0x1044] select channel by service_id\n"
      "         -n <service_name> :      [BBC ONE] select channel by service_name\n";

void usage()
{
    printf("%s", usage_str);
}

int main(int argc, char **argv)
{
    int adapter = 0, frontend = 0, demux = 0;
    int scan = 0;
    int frequency = 505833333;
    int tune = 0;
    int promiscuous = 0;
    int lcn = 1;
    int service_id = -1;
    char service_name[0x100] = "";
    char filename[0x40] = "channels.xml";

    int xmlerr;
    int frontend_fd;
    struct dvb_frontend_parameters frontend_param;
    int whole_fd = 0;
    int pat_fd = 0, sdt_fd = 0, pmt_fd = 0;
    int video_fd = 0, audio_fd = 0, pcr_fd = 0;
    unsigned char **patbuf, **nitbuf, **sdtbuf, **pmtbuf;// **eitbuf;
    int num_sections;
    int lock;
    int next_lcn = 0;
    int i,j,s,t;

    int lcnpipe;
    FILE *tv_scan;
    int rlen;
    char rbuf[0x100];
    char wbuf[0x100];

    for(i=1; i<argc; i++) {
        if (argv[i][0] == '-') {
            switch (argv[i][1]) {
                case 'a':
                    adapter = strtoul(argv[++i], NULL, 0);
                    break;
                case 'f':
                    frontend = strtoul(argv[++i], NULL, 0);
                    break;
                case 'd':
                    demux = strtoul(argv[++i], NULL, 0);
                    break;

                case 'c':
                    sscanf(argv[++i],"%[^]]", filename);
                    break;

                case 's':
                    scan ^= 1;
                    break;
                case 'e':
                    frequency = strtoul(argv[++i], NULL, 0);
                    break;
                case 't':
                    tune ^= 1;
                    break;
                case 'p':
                    promiscuous ^= 1;
                    break;
                case 'l':
                    lcn = strtoul(argv[++i], NULL, 0);
                    service_id = -1;
                    service_name[0] = '\0';
                    break;
                case 'i':
                    lcn = -1;
                    service_id = strtoul(argv[++i], NULL, 16);
                    service_name[0] = '\0';
                    break;
                case 'n':
                    lcn = -1;
                    service_id = -1;
                    sscanf(argv[++i],"%[^]]", service_name);
                    break;

                case 'v':
                    debug = strtoul(argv[++i], NULL, 0);
                    break;
                case '?' :
                default  :
                    usage();
                    return 1;
            }
        }
    }
    printf("\n");
    printf("tv_scan v0.04\n");
    printf("=============\n");
    printf("\n");
    printf("adapter      = %d\n", adapter);
    printf("frontend     = %d\n", frontend);
    printf("demux        = %d\n", demux);
    printf("channel_list = %s\n", filename);
    printf("scan         = %d\n", scan);
    printf("frequency    = %d\n", frequency);
    printf("tune         = %d\n", tune);
    printf("promiscuous  = %d\n", promiscuous);
    printf("lcn          = %d\n", lcn);
    printf("service_id   = %04x\n", service_id);
    printf("service_name = %s\n", service_name);
    printf("\n");

    signal(SIGINT,do_abort);

    mkfifo("/tmp/lcn", 0666);
    lcnpipe = open("/tmp/lcn", O_RDONLY | O_NONBLOCK);

    init_transponder();
    xmlerr = import_transponder(filename);
//    if (debug) dump_transponder(NULL);
    if (debug) dump_transponder("channels1.log");

    //set the fake_lcn to > the lowest lcn that is currently in use
    for (t=0; t<num_transponders; t++) {
        for (s=0; s<transponder[t].num_services; s++) {
            if (transponder[t].service[s].lcn > fake_lcn) fake_lcn = transponder[t].service[s].lcn + 1;
        }
    }

    snprintf(FRONTEND_DEV, sizeof(FRONTEND_DEV), "/dev/dvb/adapter%i/frontend%i", adapter, frontend);
    snprintf(DEMUX_DEV, sizeof(DEMUX_DEV), "/dev/dvb/adapter%i/demux%i", adapter, demux);
    snprintf(DVR_DEV, sizeof(DVR_DEV), "/dev/dvb/adapter%i/dvr%i", adapter, demux);

    //open frontend
    if ((frontend_fd = open(FRONTEND_DEV, O_RDWR)) < 0) {
        printf ("\nfailed opening '%s'\n", FRONTEND_DEV);
        return -1;
    }

    while(!timeout_flag) {

        if (scan) {
            //tune to initial channel
            if (xmlerr == 0) {
                //if .xml exists, tune to first channel in channels.xml
                t = 0;
                frontend_param.frequency                    = transponder[t].frequency;
                frontend_param.inversion                    = INVERSION_OFF;
                frontend_param.u.ofdm.bandwidth             = (transponder[t].bandwidth == 0) ? BANDWIDTH_8_MHZ : (transponder[t].bandwidth == 1) ? BANDWIDTH_7_MHZ : BANDWIDTH_6_MHZ;
                frontend_param.u.ofdm.code_rate_HP          = (transponder[t].fec_high == 0) ? FEC_1_2 : (transponder[t].fec_high == 1) ? FEC_2_3 : (transponder[t].fec_high == 2) ? FEC_3_4 : (transponder[t].fec_high == 3) ? FEC_5_6 : (transponder[t].fec_high == 4) ? FEC_7_8 : FEC_AUTO;
                frontend_param.u.ofdm.code_rate_LP          = (transponder[t].fec_low  == 0) ? FEC_1_2 : (transponder[t].fec_low  == 1) ? FEC_2_3 : (transponder[t].fec_low  == 2) ? FEC_3_4 : (transponder[t].fec_low  == 3) ? FEC_5_6 : (transponder[t].fec_low  == 4) ? FEC_7_8 : FEC_AUTO;
                frontend_param.u.ofdm.constellation         = (transponder[t].modulation == 1) ? QAM_16 : (transponder[t].modulation == 2) ? QAM_64 : QAM_AUTO;
                frontend_param.u.ofdm.transmission_mode     = (transponder[t].transmission_mode == 0) ? TRANSMISSION_MODE_2K : (transponder[t].transmission_mode == 1) ? TRANSMISSION_MODE_8K : TRANSMISSION_MODE_AUTO;
                frontend_param.u.ofdm.guard_interval        = (transponder[t].guard_interval == 0) ? GUARD_INTERVAL_1_32 : (transponder[t].guard_interval == 1) ? GUARD_INTERVAL_1_16 : (transponder[t].guard_interval == 2) ? GUARD_INTERVAL_1_8 : (transponder[t].guard_interval == 3) ? GUARD_INTERVAL_1_4 : GUARD_INTERVAL_AUTO;
                frontend_param.u.ofdm.hierarchy_information = (transponder[t].hierarchy == 0) ? HIERARCHY_NONE : (transponder[t].hierarchy == 1) ? HIERARCHY_1 : (transponder[t].hierarchy == 2) ? HIERARCHY_2 : (transponder[t].hierarchy == 3) ? HIERARCHY_4 : HIERARCHY_AUTO;
            }
            else {
                //else pick some defaults
                frontend_param.frequency                    = frequency;
                frontend_param.inversion                    = INVERSION_OFF;
                frontend_param.u.ofdm.bandwidth             = BANDWIDTH_8_MHZ;
                frontend_param.u.ofdm.code_rate_HP          = FEC_3_4;                  //FEC_2_3    !!! default may not work for itv ...
                frontend_param.u.ofdm.code_rate_LP          = 0;
                frontend_param.u.ofdm.constellation         = QAM_16;                   //QAM_64
                frontend_param.u.ofdm.transmission_mode     = TRANSMISSION_MODE_2K;
                frontend_param.u.ofdm.guard_interval        = GUARD_INTERVAL_1_32;
                frontend_param.u.ofdm.hierarchy_information = HIERARCHY_NONE;
            }
            printf("\n");
            printf("tuning to: %d ...\n", frontend_param.frequency);
            if (setup_frontend(frontend_fd, &frontend_param) < 0) {
                close(frontend_fd);
                return -1;
            }
            lock = check_frontend(frontend_fd);
            if (!lock) {
                printf("failed to tune to %d\n", frontend_param.frequency);
                close(frontend_fd);
                return -1;
            }
            printf("tuned OK to %d\n", frontend_param.frequency);

            //get the frequency parameters from the nit
            nitbuf = get_section(DEMUX_DEV, "network_id         ", 0x0010, 0x40, 0xff, &num_sections);
            for (i=0; i<num_sections; i++) {
                get_nit_info(nitbuf[i]);
#ifdef DEBUG
                if (debug) dump_nit(nitbuf[i]);
#endif
            }
            for (i=0; i<num_sections; i++) {
                if (nitbuf[i]) free(nitbuf[i]);
            }
            if (nitbuf) free(nitbuf);


            t = 0;
            do {
                //the madmen who created the linux dvb api decided not to use the same enums as the dvb standard ...
                frontend_param.frequency                    = transponder[t].frequency;
                frontend_param.inversion                    = INVERSION_OFF;
                frontend_param.u.ofdm.bandwidth             = (transponder[t].bandwidth == 0) ? BANDWIDTH_8_MHZ : (transponder[t].bandwidth == 1) ? BANDWIDTH_7_MHZ : BANDWIDTH_6_MHZ;
                frontend_param.u.ofdm.code_rate_HP          = (transponder[t].fec_high == 0) ? FEC_1_2 : (transponder[t].fec_high == 1) ? FEC_2_3 : (transponder[t].fec_high == 2) ? FEC_3_4 : (transponder[t].fec_high == 3) ? FEC_5_6 : (transponder[t].fec_high == 4) ? FEC_7_8 : FEC_AUTO;
                frontend_param.u.ofdm.code_rate_LP          = (transponder[t].fec_low  == 0) ? FEC_1_2 : (transponder[t].fec_low  == 1) ? FEC_2_3 : (transponder[t].fec_low  == 2) ? FEC_3_4 : (transponder[t].fec_low  == 3) ? FEC_5_6 : (transponder[t].fec_low  == 4) ? FEC_7_8 : FEC_AUTO;
                frontend_param.u.ofdm.constellation         = (transponder[t].modulation == 1) ? QAM_16 : (transponder[t].modulation == 2) ? QAM_64 : QAM_AUTO;
                frontend_param.u.ofdm.transmission_mode     = (transponder[t].transmission_mode == 0) ? TRANSMISSION_MODE_2K : (transponder[t].transmission_mode == 1) ? TRANSMISSION_MODE_8K : TRANSMISSION_MODE_AUTO;
                frontend_param.u.ofdm.guard_interval        = (transponder[t].guard_interval == 0) ? GUARD_INTERVAL_1_32 : (transponder[t].guard_interval == 1) ? GUARD_INTERVAL_1_16 : (transponder[t].guard_interval == 2) ? GUARD_INTERVAL_1_8 : (transponder[t].guard_interval == 3) ? GUARD_INTERVAL_1_4 : GUARD_INTERVAL_AUTO;
                frontend_param.u.ofdm.hierarchy_information = (transponder[t].hierarchy == 0) ? HIERARCHY_NONE : (transponder[t].hierarchy == 1) ? HIERARCHY_1 : (transponder[t].hierarchy == 2) ? HIERARCHY_2 : (transponder[t].hierarchy == 3) ? HIERARCHY_4 : HIERARCHY_AUTO;

                //tune to channel
                printf("\n");
                printf("tuning to: %d ...\n", frontend_param.frequency);
                if (setup_frontend(frontend_fd, &frontend_param) < 0) {
                    close(frontend_fd);
                    return -1;
                }
                lock = check_frontend(frontend_fd);
                if (!lock) {
                    printf("failed to tune to %d\n", frontend_param.frequency);
                    close(frontend_fd);
                    return -1;
                }
                printf("tuned OK to %d\n", frontend_param.frequency);


                //get the pmt sids, pids from the pat
                patbuf = get_section(DEMUX_DEV, "transport_stream_id", 0x0000, 0x00, 0xff, &num_sections);
                if (patbuf) {
                get_pat_info(patbuf[0]);
#ifdef DEBUG
                if (debug) dump_pat(patbuf[0]);
#endif
                for (i=0; i<num_sections; i++) {
                    if (patbuf[i]) free(patbuf[i]);
                }
                if (patbuf) free(patbuf);
                }


                //get the channel name from the sdt
                sdtbuf = get_section(DEMUX_DEV, "transport_stream_id", 0x0011, 0x42, 0xff, &num_sections);
                if (sdtbuf) {
                get_sdt_info(sdtbuf[0]);
#ifdef DEBUG
                if (debug) dump_sdt(sdtbuf[0]);
#endif
                for (i=0; i<num_sections; i++) {
                    if (sdtbuf[i]) free(sdtbuf[i]);
                }
                if (sdtbuf) free(sdtbuf);
                }


                //get a list of pids from the pmt
                for (s=0; s<transponder[t].num_services; s++) {
                    if (transponder[t].service[s].pmt_pid == 0) {
                        close(frontend_fd);
                        return -1;
                    }
                    pmtbuf = get_section(DEMUX_DEV, "service_id         ", transponder[t].service[s].pmt_pid, 0x02, 0xff, &num_sections);
                    if (pmtbuf) {
                    get_pmt_info(pmtbuf[0], t);
#ifdef DEBUG
                    if (debug) dump_pmt(pmtbuf[0], t);
#endif
                    for (i=0; i<num_sections; i++) {
                        if (pmtbuf[i]) free(pmtbuf[i]);
                    }
                    if (pmtbuf) free(pmtbuf);
                    }
                }

                t++;
            }  while (t<num_transponders);

            printf("\n");
            printf("found %d transponders\n", num_transponders);

            //save to file
            if (debug) dump_transponder("channels.log");
            export_transponder(filename);

            //write to /tmp/tv_scan
            tv_scan = fopen("/tmp/tv_scan", "wt");
            if (!tv_scan) {
                printf("failed opening '%s'\n", "/tmp/tv_scan");
                return -1;
            }
            sprintf(wbuf, "scan complete\n");
            fwrite(wbuf, 1, strlen(wbuf), tv_scan);
            fflush(tv_scan);
            fclose(tv_scan);
            scan = 0;
        }



        if (tune) {
            for (t=0; t<num_transponders; t++) {
                for (s=0; s<transponder[t].num_services; s++) {
                    if (lcn > 0) {
                        //match lcn with transponder,service
                        if (transponder[t].service[s].lcn == lcn) {
                            goto found_lcn;
                        }
                    }
                    else if (service_id >= 0) {
                        //match service_id with transponder,service
                        if (transponder[t].service[s].service_id == service_id) {
                            goto found_lcn;
                        }
                    }
                    else {
                        //match service_name with transponder,service
                        if (strncasecmp(transponder[t].service[s].service_name, service_name, strlen(service_name)) == 0) {
                            goto found_lcn;
                        }
                    }
                }
            }

            t = 0;
            s = 0;
found_lcn:

            //tune to selected mux, + update
            printf("\n");
            printf("tuning to: %d ...\n", transponder[t].frequency);
            printf("  frequency             = %d\n",      transponder[t].frequency);
            printf("  inversion             = %d (%s)\n", INVERSION_OFF,                        "INVERSION_OFF");
            printf("  bandwidth             = %d (%s)\n", transponder[t].bandwidth,             dvb_enumtype(bandwidth_str,         transponder[t].bandwidth));
            printf("  code_rate_HP          = %d (%s)\n", transponder[t].fec_high,              dvb_enumtype(code_rate_str,         transponder[t].fec_high));
            printf("  code_rate_LP          = %d (%s)\n", transponder[t].fec_low,               dvb_enumtype(code_rate_str,         transponder[t].fec_low));
            printf("  constellation         = %d (%s)\n", transponder[t].modulation,            dvb_enumtype(constellation_str,     transponder[t].modulation));
            printf("  transmission_mode     = %d (%s)\n", transponder[t].transmission_mode,     dvb_enumtype(transmission_mode_str, transponder[t].transmission_mode));
            printf("  guard_interval        = %d (%s)\n", transponder[t].guard_interval,        dvb_enumtype(guard_interval_str,    transponder[t].guard_interval));
            printf("  hierarchy             = %d (%s)\n", transponder[t].hierarchy,             dvb_enumtype(hierarchy_str,         transponder[t].hierarchy));
            //the madmen who created the linux dvb api decided not to use the same enums as the dvb standard ...
            frontend_param.frequency                    = transponder[t].frequency;
            frontend_param.inversion                    = INVERSION_OFF;
            frontend_param.u.ofdm.bandwidth             = (transponder[t].bandwidth == 0) ? BANDWIDTH_8_MHZ : (transponder[t].bandwidth == 1) ? BANDWIDTH_7_MHZ : BANDWIDTH_6_MHZ;
            frontend_param.u.ofdm.code_rate_HP          = (transponder[t].fec_high == 0) ? FEC_1_2 : (transponder[t].fec_high == 1) ? FEC_2_3 : (transponder[t].fec_high == 2) ? FEC_3_4 : (transponder[t].fec_high == 3) ? FEC_5_6 : (transponder[t].fec_high == 4) ? FEC_7_8 : FEC_AUTO;
            frontend_param.u.ofdm.code_rate_LP          = (transponder[t].fec_low  == 0) ? FEC_1_2 : (transponder[t].fec_low  == 1) ? FEC_2_3 : (transponder[t].fec_low  == 2) ? FEC_3_4 : (transponder[t].fec_low  == 3) ? FEC_5_6 : (transponder[t].fec_low  == 4) ? FEC_7_8 : FEC_AUTO;
            frontend_param.u.ofdm.constellation         = (transponder[t].modulation == 1) ? QAM_16 : (transponder[t].modulation == 2) ? QAM_64 : QAM_AUTO;
            frontend_param.u.ofdm.transmission_mode     = (transponder[t].transmission_mode == 0) ? TRANSMISSION_MODE_2K : (transponder[t].transmission_mode == 1) ? TRANSMISSION_MODE_8K : TRANSMISSION_MODE_AUTO;
            frontend_param.u.ofdm.guard_interval        = (transponder[t].guard_interval == 0) ? GUARD_INTERVAL_1_32 : (transponder[t].guard_interval == 1) ? GUARD_INTERVAL_1_16 : (transponder[t].guard_interval == 2) ? GUARD_INTERVAL_1_8 : (transponder[t].guard_interval == 3) ? GUARD_INTERVAL_1_4 : GUARD_INTERVAL_AUTO;
            frontend_param.u.ofdm.hierarchy_information = (transponder[t].hierarchy == 0) ? HIERARCHY_NONE : (transponder[t].hierarchy == 1) ? HIERARCHY_1 : (transponder[t].hierarchy == 2) ? HIERARCHY_2 : (transponder[t].hierarchy == 3) ? HIERARCHY_4 : HIERARCHY_AUTO;
            //tune to channel
            if (setup_frontend(frontend_fd, &frontend_param) < 0) {
                close(frontend_fd);
                return -1;
            }
            lock = check_frontend(frontend_fd);
            if (!lock) {
                printf("failed to tune to %d\n", frontend_param.frequency);
                close(frontend_fd);
                return -1;
            }
            printf("tuned OK to %d\n", frontend_param.frequency);


            //get the pmt sids, pids from the pat
            patbuf = get_section(DEMUX_DEV, "transport_stream_id", 0x0000, 0x00, 0xff, &num_sections);
            get_pat_info(patbuf[0]);
#ifdef DEBUG
            if (debug) dump_pat(patbuf[0]);
#endif
            for (i=0; i<num_sections; i++) {
                if (patbuf[i]) free(patbuf[i]);
            }
            if (patbuf) free(patbuf);


            //get the channel name from the sdt
            sdtbuf = get_section(DEMUX_DEV, "transport_stream_id", 0x0011, 0x42, 0xff, &num_sections);
            get_sdt_info(sdtbuf[0]);
#ifdef DEBUG
            if (debug) dump_sdt(sdtbuf[0]);
#endif
            for (i=0; i<num_sections; i++) {
                if (sdtbuf[i]) free(sdtbuf[i]);
            }
            if (sdtbuf) free(sdtbuf);


            //get a list of pids from the pmt
            if (transponder[t].service[s].pmt_pid == 0) {
                close(frontend_fd);
                return -1;
            }
            pmtbuf = get_section(DEMUX_DEV, "service_id         ", transponder[t].service[s].pmt_pid, 0x02, 0xff, &num_sections);
            get_pmt_info(pmtbuf[0], t);
#ifdef DEBUG
            if (debug) dump_pmt(pmtbuf[0], t);
#endif
            for (i=0; i<num_sections; i++) {
                if (pmtbuf[i]) free(pmtbuf[i]);
            }
            if (pmtbuf) free(pmtbuf);



            //add dvr filter
            printf("pat_pid   = %04x\n", 0x0000);
            printf("sdt_pid   = %04x\n", 0x0011);
            printf("pmt_pid   = %04x\n", transponder[t].service[s].pmt_pid);
            printf("video_pid = %04x\n", transponder[t].service[s].video_pid);
            printf("audio_pid = %04x\n", transponder[t].service[s].audio_pid);
            printf("pcr_pid   = %04x\n", transponder[t].service[s].pcr_pid);

            if (promiscuous) {
                //open all pids - do all cards support this?
                if ((whole_fd = open(DEMUX_DEV, O_RDWR)) < 0) {
                    printf("failed opening '%s'\n", DEMUX_DEV);
                    return -1;
                }
	            if (set_pesfilter(whole_fd, 0x2000, DMX_PES_OTHER, 1) < 0) {
	            	return -1;
	            }
            } else {
                //open pat, sdt, pmt, vid, aud, pcr pids
                if ((pat_fd = open(DEMUX_DEV, O_RDWR)) < 0) {
                    printf("failed opening '%s'\n", DEMUX_DEV);
                    return -1;
                }
	            if (set_pesfilter(pat_fd, 0x0000, DMX_PES_OTHER, 1) < 0) {
	            	return -1;
	            }

                if ((sdt_fd = open(DEMUX_DEV, O_RDWR)) < 0) {
                    printf("failed opening '%s'\n", DEMUX_DEV);
                    return -1;
                }
    	        if (set_pesfilter(sdt_fd, 0x0011, DMX_PES_OTHER, 1) < 0) {
    	        	return -1;
    	        }

    	        if ((transponder[t].service[s].pmt_pid!=0x1fff)&&
    	            (transponder[t].service[s].pmt_pid!=0xffffffff)) {
                    if ((pmt_fd = open(DEMUX_DEV, O_RDWR)) < 0) {
                        printf("failed opening '%s'\n", DEMUX_DEV);
                        return -1;
                    }
    	            if (set_pesfilter(pmt_fd, transponder[t].service[s].pmt_pid, DMX_PES_OTHER, 1) < 0) {
    	            	return -1;
    	            }
    	        }

    	        if ((transponder[t].service[s].video_pid!=0x1fff)&&
                    (transponder[t].service[s].video_pid!=0xffffffff)) {
                    if ((video_fd = open(DEMUX_DEV, O_RDWR)) < 0) {
                        printf("failed opening '%s'\n", DEMUX_DEV);
                        return -1;
                    }
    	            if (set_pesfilter(video_fd, transponder[t].service[s].video_pid, DMX_PES_VIDEO, 1) < 0) {
    	            	return -1;
    	            }
    	        }

    	        if ((transponder[t].service[s].audio_pid!=transponder[t].service[s].video_pid)&&
    	            (transponder[t].service[s].audio_pid!=0x1fff)&&
    	            (transponder[t].service[s].audio_pid!=0xffffffff)) {
    	            if ((audio_fd = open(DEMUX_DEV, O_RDWR)) < 0) {
                        printf("failed opening '%s'\n", DEMUX_DEV);
                        return -1;
                    }
    	            if (set_pesfilter(audio_fd, transponder[t].service[s].audio_pid, DMX_PES_AUDIO, 1) < 0) {
    	            	return -1;
    	            }
    	        }

    	        if ((transponder[t].service[s].pcr_pid!=transponder[t].service[s].video_pid)&&
    	            (transponder[t].service[s].pcr_pid!=transponder[t].service[s].audio_pid)&&
    	            (transponder[t].service[s].pcr_pid!=0x1fff)&&
    	            (transponder[t].service[s].pcr_pid!=0xffffffff)) {
    	            if ((pcr_fd = open(DEMUX_DEV, O_RDWR)) < 0) {
                        printf("failed opening '%s'\n", DEMUX_DEV);
                        return -1;
                    }
    	            if (set_pesfilter(pcr_fd, transponder[t].service[s].pcr_pid, DMX_PES_OTHER, 1) < 0) {
    	            	return -1;
    	            }
    	        }
            }

            //write lcn, service_id, service_name to /tmp/tv_scan
            tv_scan = fopen("/tmp/tv_scan", "wt");
            if (!tv_scan) {
                printf("failed opening '%s'\n", "/tmp/tv_scan");
                return -1;
            }
            sprintf(wbuf, "tuned OK to %d. %s\n", transponder[t].service[s].lcn, transponder[t].service[s].service_name);
            fwrite(wbuf, 1, strlen(wbuf), tv_scan);
            fflush(tv_scan);
            fclose(tv_scan);

            //wait
            while (!timeout_flag) {
                //could retune by lcn, frequency(transponder), service_id
                rlen = read(lcnpipe, rbuf, 0x100);
                if (rlen > 0) {
                    next_lcn = strtoul(rbuf, NULL, 10);
                    printf("retune to lcn: %d\n", next_lcn);
                    break;
                }
                usleep(100000);
            }

            if (!timeout_flag) {
                //write lcn, service_id, service_name to /tmp/tv_scan
                tv_scan = fopen("/tmp/tv_scan", "wt");
                if (!tv_scan) {
                    printf("failed opening '%s'\n", "/tmp/tv_scan");
                    return -1;
                }
                if (next_lcn) {
                    lcn = next_lcn;
                    sprintf(wbuf, "tuning to %d.\n", lcn);
                }
                else {
                    sprintf(wbuf, "scanning ...\n");
                    scan = 1;
                }
                fwrite(wbuf, 1, strlen(wbuf), tv_scan);
                fflush(tv_scan);
                fclose(tv_scan);
            }

            if (whole_fd) close(whole_fd);
            if (pat_fd) close(pat_fd);
            if (sdt_fd) close(sdt_fd);
            if (pmt_fd) close(pmt_fd);
            if (video_fd) close(video_fd);
            if (audio_fd) close(audio_fd);
            if (pcr_fd) close(pcr_fd);
        }
    }

    close(lcnpipe);
    //remove /tmp/lcn
    //remove /tmp/tv_scan
    close_transponder();
    close(frontend_fd);
    return 0;
}
